(*****************************************************************)
(*  Instances for mathcomp                                       *)
(*                                                               *)
(* Copyright (c) 2017 Frédéric Besson <frederic.besson@inria.fr> *)
(* Copyright (c) 2017 Laurent Théry   <laurent.thery@inria.fr>   *)
(*****************************************************************)

Require Import ZArith Lia.
Require Import FOLib FOCompil FOReify.

From mathcomp Require Import all_ssreflect.

Lemma dvdn_Zdiv x y :
  (x %| y) -> (Z.of_nat x | Z.of_nat y)%Z.
Proof.
case/dvdnP=> z ->.
exists (Z.of_nat z).
by rewrite Nat2Z.inj_mul.
Qed.

Lemma Zdiv_dvdn x y :
  (Z.of_nat x | Z.of_nat y)%Z -> (x %| y).
Proof.
case; case=> [|p|p H]; first by case: y.
rewrite -(Z2Nat.id (Z.pos p)) // -Nat2Z.inj_mul => /Nat2Z.inj->.
  by apply/dvdnP; exists (Z.to_nat (Z.pos p)).
have := Zle_0_nat y.
have := Zle_0_nat x.
rewrite -{2}(Nat2Z.id x) H; case: (Z.of_nat x) H => //=.
by case: y.
Qed.

Lemma gcdn_gcd : forall x y, 
    Z.of_nat (gcdn x y) = Z.gcd (Z.of_nat x) (Z.of_nat y).
Proof.
move=> x y.
apply: esym.
apply: Z.gcd_unique; try lia.
- by apply/dvdn_Zdiv/dvdn_gcdl.
- by apply/dvdn_Zdiv/dvdn_gcdr.
have F p : 
  (Z.pos p | Z.of_nat x)%Z ->
  (Z.pos p | Z.of_nat y)%Z -> (Z.pos p | Z.of_nat (gcdn x y))%Z.
  rewrite -(Z2Nat.id (Z.pos p)) // => /Zdiv_dvdn H1 /Zdiv_dvdn H2.
  by apply: dvdn_Zdiv; rewrite dvdn_gcd H1.
case => //= [/(Zdiv_dvdn 0) H1 /(Zdiv_dvdn 0) H2|p].
  by apply: (dvdn_Zdiv 0); rewrite dvdn_gcd H1.
move=> /Znumtheory.Zdivide_opp_l /= H1 /Znumtheory.Zdivide_opp_l /= H2.
by apply/(Znumtheory.Zdivide_opp_l (Z.pos p))/F.
Qed.

Lemma lcmn_lcm : forall x y, 
  Z.of_nat (lcmn x y) = Z.lcm (Z.of_nat x) (Z.of_nat y).
Proof.
move=> x y.
apply: esym.
apply: Z.lcm_unique; try lia.
- by apply/dvdn_Zdiv/dvdn_lcml.
- by apply/dvdn_Zdiv/dvdn_lcmr.
have F p : 
  (Z.of_nat x | Z.pos p)%Z ->
  (Z.of_nat y | Z.pos p)%Z -> (Z.of_nat (lcmn x y) | Zpos p)%Z.
  rewrite -(Z2Nat.id (Z.pos p)) // => /Zdiv_dvdn H1 /Zdiv_dvdn H2.
  by apply: dvdn_Zdiv; rewrite dvdn_lcm H1.
case => //= [|p]; first by move=> *; apply: Z.divide_0_r.
move=> /Znumtheory.Zdivide_opp_r /= H1.
move=> /Znumtheory.Zdivide_opp_r /= H2.
by apply/(Znumtheory.Zdivide_opp_r _ (Z.pos p))/F.
Qed.

(** ***)


Instance ZType : @TypeDecl.t Z.
Proof.
  apply TypeDecl.Mk with (equi := @eq Z) (type_cstr := None).
  * exact Z0.
  * typeclasses eauto.  * intros ; auto.
    discriminate.
Defined.

Instance natType : @TypeDecl.t nat.
apply TypeDecl.Mk with (equi := @eq nat) (type_cstr := Some (fun x => 0 <= x = true)).
* exact O.
* typeclasses eauto.
*  intros. inv H.
   destruct t ; simpl; auto.
Defined.

Instance boolType : @TypeDecl.t bool.
apply TypeDecl.Mk with (equi := @eq bool) (type_cstr := Some (fun x => x = true \/ x = false)).
* exact true.
* typeclasses eauto.
*  intros. inv H.
   destruct t ; tauto.
Defined.

Instance injNatZ : DeclInj.t nat Z. 
Proof.
apply (DeclInj.Mk nat Z Z.of_nat) ; simpl.
* split.
  congruence.
  apply Znat.Nat2Z.inj ; auto.
Defined.

Instance injZZ : DeclInj.t Z Z.
Proof.
  apply (DeclInj.Mk Z Z (fun x => x)).
  simpl ; auto.
  tauto.
Defined.

Instance injBoolProp : DeclInj.t bool Prop. 
Proof.
apply (DeclInj.Mk bool Prop is_true) ; simpl.
split; intros.
- rewrite H; tauto.
- destruct x ; destruct y ; unfold is_true in *;
    intuition congruence.
Defined.

Instance injPropProp : DeclInj.t Prop Prop.
Proof.
  apply (DeclInj.Mk Prop Prop (fun x =>x)) ; simpl.
  tauto.
Defined.  



Instance Injis_true : Inj.t _ is_true.
Proof.
  make_inj (fun (x:Prop) => x).
Defined.

Instance InjAdd  : Inj.t  _ plus. 
Proof.
  make_inj Zplus.
  apply Znat.Nat2Z.inj_add.
Defined.

Instance InjAddn  : Inj.t  _ addn. 
Proof.
  make_inj Zplus.
  apply Znat.Nat2Z.inj_add.
Defined.

Instance InjMinus  : Inj.t  _ minus. 
Proof.
  make_inj (fun n m => Z.max 0 (Zminus n m)).
  apply Znat.Nat2Z.inj_sub_max.
Defined.

Instance InjSubn  : Inj.t  _ subn. 
Proof.
  make_inj (fun n m => Z.max 0 (Zminus n m)).
  apply Znat.Nat2Z.inj_sub_max.
Defined.


Instance InjMult  : Inj.t  _ mult. 
Proof.
  make_inj Zmult.
  apply Znat.Nat2Z.inj_mul.
Defined.

Instance InjMuln  : Inj.t  _ muln. 
Proof.
  make_inj Zmult.
  apply Znat.Nat2Z.inj_mul.
Defined.


Instance InjMax  : Inj.t  _ max. 
Proof.
  make_inj Z.max.
  apply Znat.Nat2Z.inj_max.
Defined.

Instance InjLe : Inj.t _ leq. 
Proof.
  make_inj Z.le.
  generalize (Znat.Nat2Z.inj_le a0 a1).
  intro.
  generalize (@leP a0 a1).
  intro HH. inv HH.
  - intuition.
  - unfold is_true. intuition congruence.
Defined.

Instance InjExpn : Inj.t _ expn.
Proof.
  make_inj Z.pow.
  induction a1.
  - auto.
  - rewrite expnS.
    rewrite Znat.Nat2Z.inj_succ.
    unfold Z.succ.
    rewrite Zpower.Zpower_exp.
    rewrite Znat.Nat2Z.inj_mul.
    + rewrite IHa1.
      ring.
    + change 0%Z with (Z.of_nat O).
      apply Znat.inj_ge.
      apply le_0_n.
      compute. congruence.
Defined.

Instance InjEqn : Inj.t _ eqn. 
Proof.
  make_inj (@eq Z).
  destruct (@eqnP a0 a1).
  - rewrite e. 
    split; auto.
  - unfold is_true.
    split ; try congruence.
    intro.
    apply Znat.Nat2Z.inj in H. congruence.
Defined.
    
Instance InjO      : Inj.t _ O.      
Proof.
  make_inj Z0.
Defined.


Instance InjS      : Inj.t _ S.      
Proof.
  make_inj Z.succ.
  apply Znat.Zpos_P_of_succ_nat.
Defined.

Instance Injpred : Inj.t _  predn.      
Proof.
  make_inj (fun x => Z.max 0 (x - 1)).
  rewrite Znat.Nat2Z.inj_pred_max.
  reflexivity.
Defined.

Instance InjBTrue : Inj.t _ true. 
Proof.
  make_inj True.
  unfold is_true. tauto.
Defined.

Instance InjBFalse : Inj.t _ false. 
Proof.
  make_inj False.
  unfold is_true. intuition congruence.
Defined.

Instance InjTrue  : Inj.t _ True.   
Proof.
  make_inj True.
Defined.

Instance InjFalse  : Inj.t _ False.   
Proof.
  make_inj False.
Defined.

Instance InjZof_nat : Inj.t _ Z.of_nat.
Proof.
  make_inj (fun x:Z => x).
Defined.

Instance InjNot  : Inj.t _ not.   
Proof.
  make_inj not.
Defined.

Instance InjOr  : Inj.t _ or.   
Proof.
  make_inj or.
Defined.

Instance InjAndb  : Inj.t _ andb.   
Proof.
  make_inj and.
  destruct a0 ; destruct a1 ; simpl;
    intuition congruence.
Defined.

Instance InjOrb  : Inj.t _ orb.   
Proof.
  make_inj or.
  destruct a0 ; destruct a1 ; simpl;
    intuition congruence.
Defined.

Instance InjNegb  : Inj.t _ negb.   
Proof.
  make_inj not.
  destruct a0 ; simpl;
    intuition congruence.
Defined.

Definition z_of_bool (b : bool) : Z :=
  if b then 1%Z else 0%Z.

Definition uphalfZ (z :Z) := (z_of_bool (Z.odd z) + Z.div z  2)%Z.

Lemma inj_odd : forall x, 
    (odd x) = Z.odd (Z.of_nat  x).
Proof.
  induction x ; auto.
  rewrite Znat.Nat2Z.inj_succ.
  rewrite Z.odd_succ.
  simpl.
  rewrite IHx.
  rewrite Zeven.Zeven_odd_bool.
  reflexivity.
Qed.

Lemma div_div : forall n m,
    Z.of_nat (n %/ m) = Z.div (Z.of_nat n) (Z.of_nat m).
Proof.
  unfold divn.
  move => n m.
  destruct m.
  -   rewrite Zdiv.Zdiv_0_r.
      simpl. reflexivity.
  - 
  generalize (edivnP n (m.+1)).
  intro.
  destruct H.
  simpl.
  subst.
  rewrite Znat.Nat2Z.inj_add.
  rewrite Znat.Nat2Z.inj_mul.
  rewrite Zdiv.Z_div_plus_full_l.
  rewrite Z.div_small. ring.
  split.
  + apply Zorder.Zle_0_nat.
    apply Znat.Nat2Z.inj_lt.
    destruct (@ltP r (m.+1)); auto.
    destruct (@ltP 0 (m.+1)); auto.
    done.
    Lia.lia.
    Lia.lia.
Qed.

Lemma half_div2 : forall n,
    Z.of_nat n./2 = (Z.of_nat n / 2)%Z.
Proof.
  move => n.
  rewrite <- divn2.
  rewrite div_div.
  reflexivity.
Qed.

Instance InjUphalf : Inj.t _ uphalf.
Proof.
  make_inj uphalfZ.
  unfold uphalfZ.
  rewrite uphalf_half.
  rewrite Znat.Nat2Z.inj_add.
  rewrite inj_odd.
  rewrite half_div2.
  destruct (Z.odd (Z.of_nat a0)) ; reflexivity.
Defined.

Instance InjMaxn : Inj.t _ maxn.
Proof.
  make_inj Z.max.
  rewrite maxnE.
  rewrite Znat.Nat2Z.inj_add.
  rewrite Znat.Nat2Z.inj_sub_max.
  Lia.lia.
Defined.

Instance InjMinn : Inj.t _ minn.
Proof.
  make_inj Z.min.
  rewrite minnE.
  rewrite Znat.Nat2Z.inj_sub_max.
  rewrite Znat.Nat2Z.inj_sub_max.
  Lia.lia.
Defined.


Instance InjDiv : Inj.t _ divn.
Proof.
  make_inj Z.div.
  apply div_div.
Defined.

Instance InjMod : Inj.t _ modn.
make_inj (fun x y => x - x / y * y)%Z.
rewrite <- div_div.
assert ( Z.of_nat (a0 %% a1) = Z.of_nat a0 - Z.of_nat (a0 %/ a1 * a1))%Z.
{
  generalize (divn_eq  a0 a1).
  intro.
  rewrite {2}(H).
  clear H.
  rewrite Znat.Nat2Z.inj_add.
  Lia.lia.
}
rewrite H.
rewrite Znat.Nat2Z.inj_mul.
reflexivity.
Defined.

Instance InjDvdn : Inj.t  _ dvdn.
Proof.
  make_inj Z.divide.
  have := @dvdnP a0 a1.
  move => R ; destruct R.
  destruct e.
  unfold Z.divide.
  split ; auto.
  intro.
  exists (Z.of_nat x).
  rewrite H.
  rewrite Nat2Z.inj_mul.
  reflexivity.
  split.
  done.
  intro H ; destruct H.
  exfalso. apply n.
  exists (Z.to_nat x).
  clear n.
  destruct a0.
  -  simpl in H.
     rewrite Z.mul_comm in H.
     simpl in H.
     change 0%Z with (Z.of_nat 0) in H.
     apply Nat2Z.inj in H. subst ; ring.
  - 
    assert (exists k, x = Z.of_nat k).
    { exists (Z.to_nat x).
      rewrite Z2Nat.id; auto.
      have := Zle_0_nat a1 =>  A1.
      have := Zle_0_nat a0 =>  A0.
      nia.
    }
    destruct H0.
    subst.
    rewrite <- Nat2Z.inj_mul in H.
    apply Nat2Z.inj in H.
    rewrite Nat2Z.id.
    auto.
Defined.
    

Instance InjGcd : Inj.t _ gcdn.
Proof.
  make_inj Z.gcd.
  apply gcdn_gcd.
Defined.

  
Instance InjLcdn : Inj.t _ lcmn.
Proof.
  make_inj Z.lcm.
  apply lcmn_lcm.
Defined.


Instance InjHalf : Inj.t _ half.
Proof.
  make_inj (fun x => Z.div x (Z.of_nat 2)).
  apply half_div2.
Defined.

Instance InjImplb : Inj.t _ implb.
Proof.
  make_inj (fun x y : Prop=> x -> y).
  case a0, a1; simpl; intuition try congruence.
  apply H. auto.
Defined.

  
Lemma Zabs_of_nat : forall n, Z.abs (Z.of_nat n) = Z.of_nat n.
Proof.
  destruct n ; simpl.
  - auto.
  - apply f_equal.
    reflexivity.
Qed.

Instance AbstZMax : Abstract.t _ Z.max.
Proof.
  mk_abstract (fun r x y => not (x < y) /\ r = x \/ x < y /\ r = y)%Z.
  (generalize (Zmax.Zmax_spec a0 a1);
   intuition congruence).
Defined.

Instance AbstZMin : Abstract.t _ Z.min.
Proof.
  mk_abstract (fun r x y => (x <= y) /\ r = x \/ x > y /\ r = y)%Z.
  (generalize (Zmin.Zmin_spec a0 a1);
   intuition congruence).
Defined.


Instance AbstZdiv : Abstract.t _ Z.div.
Proof.
  mk_abstract 
     (fun r a b e =>   (not (b = 0%Z) -> 
                      a = b * r + e
                      /\
                      (0 <= e < b \/
                       b < e <= 0%Z)))%Z.
  (* specific proof *)
  intro.
  exists ( a0 mod a1)%Z.
  rewrite <- H.
  generalize  (Zdiv.Z_div_mod_full a0 a1).
  unfold Z.div, Z.modulo.
  destruct (Z.div_eucl a0 a1).
  unfold Zdiv.Remainder.
  intuition congruence.
Defined.
  
Instance AbstuphalfZ : Abstract.t _ uphalfZ.
Proof.
  mk_abstract ((fun r x => (x = Z.of_nat 2 * r  -1) \/ (x = Z.of_nat 2 * r))%Z).
  unfold uphalfZ.
  assert (M := Zdiv.Zmod_odd a0).
  assert (GE : (2 > 0)%Z) by Lia.lia.
  assert (EQ:= @Zdiv.Z_div_mod_eq a0 2 GE).
  rewrite M in EQ.
  destruct (Z.odd a0); rewrite /z_of_bool.
  - 
    left.
    change (a0 = 2 * r - 1)%Z.
    Lia.lia.
  - right.
    change (a0 = 2 * r)%Z.
    Lia.lia.
Defined.
